#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>

#include "mountlist.h"
#include "fsusage.h"

#define MIN_STAT 10485760
#define MAX_SORT 4096

struct dir_info {
	double usage;
	char *path;
};

typedef struct dir_info dirInfo;

dirInfo d_info[MAX_SORT];
int size = 0;

void add_info(const char *path, double usage) {
	int len = strlen(path);
	d_info[size].path = (char *) malloc(sizeof(char) * (len + 1));
	strcpy(d_info[size].path, path);
	d_info[size].usage = usage;
	size++;
}

void del_info() {
	int i = 0;
	for (i=0; i<size; i++) {
		d_info[i].usage = 0;
		free(d_info[i].path);
	}
	size = 0;
}

void swap_info(dirInfo *a, dirInfo *b) {
	double t = (*a).usage;
	(*a).usage = (*b).usage;
	(*b).usage = t;
	char *s = (*a).path;
	(*a).path = (*b).path;
	(*b).path = s;
}

void sort_info(dirInfo *array, int n) {
	float factor = 1.3;
	int g = n;
	int swapped = 1;
	while ( g>1 || swapped ) {
		g = (g>1) ? g/factor : g;
		swapped = 0;
		int i = 0;
		while ( g+i < n ) {
			if ( array[i].usage < array[g+i].usage ) {
				swap_info(&array[i], &array[g+i]);
				swapped = 1;
			}
			i++;
		}
	}
}

void format_usage(char *str, double usage) {
	double d = usage;
	int t = 0;
	while ( d >= 1024 ) {
		d /= 1024;
		t++;
	}
	sprintf(str, "%.1f", d);
	switch ( t ) {
		case 0 : strcat(str, "B"); break;
		case 1 : strcat(str, "K"); break;
		case 2 : strcat(str, "M"); break;
		case 3 : strcat(str, "G"); break;
		default : strcat(str, "T");
	}
}

void format_rate(char *str, float rate) {
	sprintf(str, "%.1f", rate);
	strcat(str, "%");
}

void print_top_info(int max) {
	int i = 0;
	for (i=0; i<max; i++) {
		if ( d_info[i].usage == 0 ) break;
		char str[128];
		format_usage(str, d_info[i].usage);
		printf("%d\t%s\t%s\n", i+1, str, d_info[i].path);
	}
}

double dir_stat(char *dir)
{
	double usage = 0;
	double dir_usage = 0;
	DIR *dp;
	struct dirent *entry;
	struct stat statbuf;
	char path[512];

	if( (dp = opendir(dir)) == NULL ) {
		return 0;
	}
	lstat(dir, &statbuf);
	__dev_t root_dev = statbuf.st_dev;
	chdir(dir);
	while( (entry = readdir(dp)) != NULL ) {
		lstat(entry->d_name,&statbuf);
		if(S_ISDIR(statbuf.st_mode)) {
			if ( statbuf.st_dev != root_dev ) continue;
			if ( !strcmp(".", entry->d_name ) || !strcmp("..", entry->d_name) ) continue;
			dir_usage = dir_stat(entry->d_name);
			getcwd(path, sizeof(path));
			if ( strcmp(path, "/") ) {
				strcat(path, "/");
			}
			strcat(path, entry->d_name);
			usage += dir_usage;

			if ( dir_usage > MIN_STAT ) {
				//printf("%d\t%u\t%s/\n", statbuf.st_dev, dir_usage, path);
				add_info(path, dir_usage);
			}
		}
		else {
			usage += statbuf.st_size;
		}
	}
	chdir("..");
	closedir(dp);
	return usage;
}

void dev_stat(char *topdir) {
	char str[128];
	int cnt = 10;
	double usage = dir_stat(topdir);
	if ( usage > MIN_STAT ) {
		add_info(topdir, usage);
	}
	format_usage(str, usage);
	printf("total accessible file size is %s.\n", str);
	format_usage(str, MIN_STAT);
	printf("%d dirs take up too much (over %s) under mount point %s\n", size, str, topdir);
	if ( size > 0 ) {
		sort_info(d_info, size);
		printf("top %d items are:\n", (cnt>size) ? size : cnt);
		print_top_info(cnt);
		del_info();
	}
}

void print_usage(const char *prog) {
	time_t now;
	struct tm ts;
	char str_m[32];
	time(&now);
	ts = *localtime(&now);
	strftime(str_m, sizeof(str_m), "%Y-%m-%d %H:%M:%S", &ts);
	printf("Do you want to know which directories eat your hard disk?\n");
	printf("eathd (version 0.1.0, build date 2010-11-19) will help you.\n");
	printf("Usage: %s [mount_point]\n", prog);
	printf("%s starts at %s\n", prog, str_m);
	printf("----------------------------------------------------------------------\n");
}

int main(int argc, char **argv) {
	double total, nonroot_total, available, available_to_root, reserved, used;
	char str_t[32], str_a[32], str_u[32], str_r[32], str_s[32];
	struct mount_entry *me;
	struct mount_entry *mount_list = read_file_system_list(true);

	print_usage(argv[0]);
	for ( me = mount_list; me; me = me->me_next ) {
		int i;
		if ( argc > 1 ) {
			bool found = false;
			for ( i = 1; i < argc; i++ ) {
				found = ( !strcmp(me->me_mountdir, argv[i]) );
				if ( found ) break;
			}
			if ( !found ) continue;
		}
		struct fs_usage fsu;
		get_fs_usage(me->me_mountdir, me->me_devname, &fsu);
		if ( fsu.fsu_blocks == 0 ) continue;
		total             = fsu.fsu_blocksize * fsu.fsu_blocks;
		available_to_root = fsu.fsu_blocksize * fsu.fsu_bfree;
		available         = fsu.fsu_blocksize * fsu.fsu_bavail;
		reserved          = fsu.fsu_blocksize * (fsu.fsu_bfree - fsu.fsu_bavail);
		used              = total - available_to_root;
		nonroot_total     = used + available;
		format_usage(str_t, total);
		format_usage(str_u, used);
		format_usage(str_a, available);
		format_usage(str_s, reserved);
		format_rate(str_r, used * 100.0 / nonroot_total);
		printf("device name: %s, device type: %s, block size: %dB\n", me->me_devname, me->me_type, (int) fsu.fsu_blocksize);
		printf("%10s%10s%10s%10s%10s   %-s\n", "Total", "Reserved", "Used", "Available", "Use%", "Mounted on");
		printf("%10s%10s%10s%10s%10s   %-s\n", str_t, str_s, str_u, str_a, str_r, me->me_mountdir);
		dev_stat(me->me_mountdir);
		printf("----------------------------------------------------------------------\n");
	}
	return 0;
}
